"""Provide easy importing for the procgen module."""

from tessellation.procgen.action import ALL_ACTIONS, Action
from tessellation.procgen.generation_result import GenerationResult
from tessellation.procgen.generator import Generator
from tessellation.procgen.tessellation_type import TessellationType



"""Random number generator based tessellation mask generator."""

from typing import Optional

import numpy as np

from tessellation.procgen import TessellationType
from tessellation.procgen.generator import Action, Generator, GenerationResult

VALID_ACTIONS = [
    Action.UP,
    Action.UP_RIGHT,
    Action.RIGHT,
    Action.DOWN,
    Action.DOWN_RIGHT,
]


class RNGGenerator(Generator):
    """Generator that uses a random number generator to generate the tesselation mask."""

    def __init__(self, seed: int = 42, side_len: int = 100):
        self.rng = np.random.default_rng(seed)
        self.side_len = side_len

    def generate(
        self,
        action_probs: Optional[list[Action]] = None,
    ) -> GenerationResult:
        """Generate a tesselation mask using a random number generator."""
        if action_probs is None:
            action_probs = [1 / len(VALID_ACTIONS)] * len(VALID_ACTIONS)

        assert len(VALID_ACTIONS) == len(action_probs)

        y_axis_mask, _ = self._generate_side(self.side_len, action_probs)
        x_axis_mask, _ = self._generate_side(self.side_len, action_probs)
        # Transpose the x_mask to make the line run along the x-axis
        x_axis_mask = x_axis_mask.T
        return GenerationResult(
            mask=x_axis_mask | y_axis_mask,
            tessellation_type=TessellationType.SQUARE_TRANSLATION,
        )

    def _generate_side(
        self, side_len: int, action_probs: list
    ) -> tuple[np.ndarray, list[Action]]:
        cursor = {"x": 0, "y": 0}
        action_list = []
        while cursor["x"] < side_len - 1:

            action = self._get_rand_action(
                actions_list=VALID_ACTIONS, action_probs=action_probs
            )
            action_list.append(action)

            if action == Action.UP:
                cursor["y"] -= 1
            elif action == Action.UP_RIGHT:
                cursor["y"] -= 1
                cursor["x"] += 1
            elif action == Action.RIGHT:
                cursor["x"] += 1
            elif action == Action.DOWN:
                cursor["y"] += 1
            elif action == Action.DOWN_RIGHT:
                cursor["y"] += 1
                cursor["x"] += 1
            else:
                raise ValueError(f"Unsupported action: {action}")

        mask = np.zeros((side_len, side_len), dtype=int)
        return self._draw_line(mask, (0, 0), action_list), action_list

    def _get_rand_action(
        self, actions_list: list[Action], action_probs: list[float]
    ) -> Action:
        """Choose a random action from the list of actions with the given probabilities."""
        return self.rng.choice(np.array(actions_list), p=action_probs)



"""Base class for tesselation generators."""

from abc import ABC

import numpy as np

from tessellation.procgen.action import Action
from tessellation.procgen.generation_result import GenerationResult


class Generator(ABC):
    """Base class for tesselation generators."""

    def generate(self) -> GenerationResult:
        """Generate a new tesselation."""
        raise NotImplementedError

    @staticmethod
    def _draw_line(
        mask: np.ndarray,
        start_point: tuple[int, int],  # (y, x)
        action_list: list[Action],
    ) -> np.ndarray:
        """Draw a line on the mask."""
        cursor = {"x": start_point[1], "y": start_point[0]}
        mask[cursor["y"], cursor["x"]] = 1
        for action in action_list:
            if action in [Action.UP_RIGHT, Action.UP]:
                cursor["y"] -= 1
            if action in [Action.DOWN_RIGHT, Action.DOWN]:
                cursor["y"] += 1
            if action in [Action.UP_RIGHT, Action.DOWN_RIGHT, Action.RIGHT]:
                cursor["x"] += 1

            if cursor["y"] >= 0:
                mask[0 : cursor["y"] + 1, cursor["x"]] = 1
            else:
                mask[cursor["y"] :, cursor["x"]] = 1

        return mask



"""Base class for tesselation generators."""

import json
from pathlib import Path
from typing import Optional, Any, Union

import numpy as np

from tessellation.procgen.tessellation_type import TessellationType


class GenerationResult:
    """Class representing the result of a tesselation generation."""

    def __init__(
        self,
        mask: np.ndarray,
        tessellation_type: TessellationType,
        metadata: Optional[dict] = None,
    ):
        self.mask = mask
        self.tessellation_type = tessellation_type
        self.metadata = metadata or {}

    def save_as_json(self, file_path: Union[Path, str]):
        """Save the generation result as a JSON file."""
        with open(file_path, "w", encoding="utf-8") as file:
            json.dump(self.to_json(), file, indent=2)

    def to_json(self) -> dict[str, Any]:
        """Return the generation result as a JSON serializable dictionary."""
        return {
            "mask": self.mask.tolist(),
            "tessellation_type": self.tessellation_type.value,
            "metadata": {**self.metadata},
        }

    @staticmethod
    def read_json(file_path: Union[Path, str]):
        """Read the generation result from a JSON file."""
        with open(file_path, "r", encoding="utf-8") as file:
            data = json.load(file)

        return GenerationResult(
            mask=np.array(data["mask"]),
            tessellation_type=TessellationType(data["tessellation_type"]),
            metadata=data["metadata"],
        )



"""Definitions for LEAP genome/phenome primitives for tessellation generation."""

import dataclasses

from leap_ec import Decoder

from tessellation.procgen import Action


@dataclasses.dataclass
class TessellationGenome:
    """Class that represents a tessellation genome."""

    actions: list[Action]
    start_point: tuple[int, int]

    def __eq__(self, other):
        return self.actions == other.actions and self.start_point == other.start_point


@dataclasses.dataclass
class TessellationPhenome:
    """Class that represents a tessellation phenome."""

    line_indices: list[tuple[int, int]]

    def __eq__(self, other):
        return self.line_indices == other.line_indices


class TessellationDecoder(Decoder):
    """Decoder for tessellation genomes."""

    def decode(
        self, genome: TessellationGenome, *args, **kwargs
    ) -> TessellationPhenome:
        """Decode a genome into a phenome."""
        cursor = {"y": genome.start_point[0], "x": genome.start_point[1]}
        line_indices = [(cursor["y"], cursor["x"])]
        for action in genome.actions:
            if action == Action.UP:
                cursor["y"] -= 1
            elif action == Action.UP_RIGHT:
                cursor["y"] -= 1
                cursor["x"] += 1
            elif action == Action.RIGHT:
                cursor["x"] += 1
            elif action == Action.DOWN:
                cursor["y"] += 1
            elif action == Action.DOWN_RIGHT:
                cursor["y"] += 1
                cursor["x"] += 1
            else:
                raise ValueError(f"Unsupported action: {action}")

            line_indices.append((cursor["y"], cursor["x"]))
        return TessellationPhenome(line_indices=line_indices)



"""Enum representing the possible actions for the generator."""

from enum import Enum, auto


class Action(Enum):
    """Enum representing the possible actions for the generator."""

    UP = auto()
    UP_RIGHT = auto()
    # UP_LEFT = auto()
    DOWN = auto()
    DOWN_RIGHT = auto()
    # DOWN_LEFT = auto()
    RIGHT = auto()
    # LEFT = auto()


ALL_ACTIONS = list(Action)



"""Methods for evolving tessellation genomes."""

from typing import Optional, Iterator, Callable

import numpy as np
from leap_ec import Individual

from tessellation.procgen import Action, ALL_ACTIONS
from tessellation.procgen.ga.genome import TessellationGenome

rng = np.random.default_rng(42)


def apply_mutation(
    mutation_fn: Callable[[Action, ...], list[Action]],
    individual_iterator: Iterator,
    fn_kwargs: Optional[dict] = None,
) -> Individual:
    """Apply mutation to each individual."""
    if fn_kwargs is None:
        fn_kwargs = {}

    while True:
        individual = next(individual_iterator)
        genome = individual.genome
        new_action_list: list[Action] = []
        for action in genome.actions:
            new_actions = mutation_fn(action, **fn_kwargs)
            new_action_list.extend(new_actions)

        individual.fitness = None  # invalidate fitness since we have new genome

        individual.genome = TessellationGenome(
            actions=new_action_list, start_point=genome.start_point
        )

        yield individual


def substitute_action(
    existing_action: Action,
    substitution_prob: float = 0.1,
    action_probs: Optional[list[float]] = None,
) -> list[Action]:
    """Substitute existing action with random action with substitution_prob."""
    if action_probs is None:
        action_probs = np.ones(len(Action)) / len(Action)

    if rng.random() < substitution_prob:
        return [rng.choice(np.array(ALL_ACTIONS), p=action_probs)]
    return [existing_action]


def insert_action(
    existing_action: Action,
    insertion_prob: float = 0.1,
    action_probs: Optional[list[float]] = None,
) -> list[Action]:
    """Insert random action with insertion_prob probability."""
    if action_probs is None:
        action_probs = np.ones(len(Action)) / len(Action)

    if rng.random() < insertion_prob:
        new_action = rng.choice(np.array(ALL_ACTIONS), p=action_probs)
        if rng.random() < 0.5:
            return [existing_action, new_action]
        return [new_action, existing_action]
    return [existing_action]


def delete_action(existing_action: Action, deletion_prob: float = 0.1) -> list[Action]:
    """Delete existing action with deletion_prob probability."""
    if rng.random() < deletion_prob:
        return []
    return [existing_action]



"""Heuristic functions for evaluating tessellation genomes."""

from tessellation.procgen.ga.genome import TessellationPhenome


### Penalties - return negative scores ###

DISQUALIFICATION_FITNESS = -100_000


def bottom_top_not_even_penalty(
    phenome: TessellationPhenome, max_diff_before_penalty: int
) -> float:
    """Check that the bottom and top sides have a relatively even number of points."""
    n_top_points = 0
    n_bottom_points = 0
    for idx in phenome.line_indices:
        y_idx = idx[0]
        if y_idx >= 0:
            n_top_points += 1
        else:
            n_bottom_points += 1
    n_points_diff = abs(n_top_points - n_bottom_points)
    if n_points_diff <= max_diff_before_penalty:
        return 0

    points_over_threshold = n_points_diff - max_diff_before_penalty
    return -1 * points_over_threshold


def duplicated_points_penalty(phenome: TessellationPhenome) -> float:
    """Returns negative heuristic value for duplicated points."""
    n_line_indices = len(phenome.line_indices)
    n_unique_line_indices = len(set(phenome.line_indices))
    return n_unique_line_indices - n_line_indices


def out_of_bounds_penalty(phenome: TessellationPhenome, side_len: int) -> float:
    """Check that the tessellation line does not go out of bounds."""
    max_x, max_y = side_len - 1, side_len - 1
    min_x, min_y = 0, -1 * side_len
    in_bounds = all(
        min_x <= x <= max_x and min_y <= y <= max_y for (y, x) in phenome.line_indices
    )
    if in_bounds:
        return 0
    return DISQUALIFICATION_FITNESS


def stray_pixels_penalty(phenome: TessellationPhenome) -> float:
    """Penalty for stray pixels in the tessellation line."""
    # Implement fitness function to check for stray pixels
    raise NotImplementedError()


def jagged_edges_penalty(phenome: TessellationPhenome) -> float:
    """Penalty for jagged edges in the tessellation line."""
    # Implement fitness function to check for jagged edges
    raise NotImplementedError()


### Rewards - return positive scores ###


def count_number_points_reward(phenome: TessellationPhenome) -> float:
    """Count the number of points in the tessellation line."""
    return len(phenome.line_indices)


def bottom_top_even_reward(
    phenome: TessellationPhenome, max_diff_before_reward: int
) -> float:
    """Check that the bottom and top sides have a relatively even number of points."""
    n_top_points = 0
    n_bottom_points = 0
    for idx in phenome.line_indices:
        y_idx = idx[0]
        if y_idx >= 0:
            n_top_points += 1
        else:
            n_bottom_points += 1

    n_points_diff = abs(n_top_points - n_bottom_points)
    if n_points_diff >= max_diff_before_reward:
        return 0

    return max_diff_before_reward - n_points_diff


def reaches_corner_to_corner_reward(phenome: TessellationPhenome) -> float:
    """Reward for tessellation line that reaches from corner to corner of mask."""
    # Implement fitness function to check that the line is connected
    raise NotImplementedError()



"""Enum representing the possible tesselation types."""

from enum import Enum


class TessellationType(str, Enum):
    """Enum representing the possible tesselation types."""

    SQUARE_TRANSLATION = "SQUARE_TRANSLATION"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tessellation',
 'tessellation.procgen',
 'tessellation.procgen.ga',
 'tessellation.procgen.rng']

package_data = \
{'': ['*']}

install_requires = \
['leap-ec>=0.8.1,<0.9.0',
 'matplotlib>=3.9.0,<4.0.0',
 'numpy>=2.0.1,<3.0.0',
 'scikit-image>=0.24.0,<0.25.0']

setup_kwargs = {
    'name': 'tessellation',
    'version': '0.0.6',
    'description': 'A service for procedurally generating tessellations.',
    'long_description': '# tessellation\nThis repository holds code to procedurally generate tessellations.\n',
    'author': "Anthony D'Achille",
    'author_email': 'adachille15@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/adachille/tessellation',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



"""Genetic Algorithm based tessellation generator."""

from typing import Callable, Optional

import numpy as np
from leap_ec import Representation, ops, probe, Individual
from leap_ec.algorithm import generational_ea

from tessellation.procgen import Generator, GenerationResult, Action, TessellationType
from tessellation.procgen.ga.genome import TessellationPhenome, TessellationDecoder
from tessellation.procgen.ga.problem import TessellationProblem, initialize_genome


class GATessellationGenerator(Generator):
    """Genetic Algorithm based tessellation generator."""

    def __init__(
        self,
        heuristic_fns: list[Callable[[TessellationPhenome], float]],
        mutation_fns: list[Callable[[Action, ...], list[Action]]],
        heuristic_fn_weights: Optional[list[float]] = None,
        side_len: int = 100,
        max_generations: int = 100,
        population_size: int = 100,
    ):
        self.side_len = side_len
        self.problem = TessellationProblem(
            heuristic_fns=heuristic_fns,
            fn_weights=heuristic_fn_weights,
            side_len=self.side_len,
        )
        self.representation = Representation(
            decoder=TessellationDecoder(),
            initialize=lambda: initialize_genome(self.problem),
        )
        self.mutation_fns = mutation_fns

        self.max_generations = max_generations
        self.population_size = population_size

        self.population = None

    def generate(self, individual_idx: int = 0) -> GenerationResult:
        """Generate a new tesselation."""
        if self.population is None:
            self.evolve()

        return self.get_generation_result(self.population[individual_idx])

    def evolve(self) -> list[GenerationResult]:
        """Run genetic algorithm and evolve the population."""
        self.population = generational_ea(
            max_generations=self.max_generations,
            pop_size=self.population_size,
            problem=self.problem,
            representation=self.representation,
            # The operator pipeline
            pipeline=[
                ops.tournament_selection,  # Select parents via tournament
                ops.clone,  # Copy them (just to be safe)
                *self.mutation_fns,  # Apply mutation functions
                # Crossover w/ 40% chance of swapping gen
                # ops.UniformCrossover(p_swap=0.4), es
                ops.evaluate,  # Evaluate fitness
                # pylint: disable=no-value-for-parameter
                ops.pool(size=self.population_size),  # Collect offspring into new pop
                probe.BestSoFarProbe(),  # Print best so far
            ],
        )
        return [
            self.get_generation_result(individual) for individual in self.population
        ]

    def get_generation_result(
        self, individual: Individual
    ) -> Optional[GenerationResult]:
        """Return the generation result for the individual."""
        genome = individual.genome
        mask = np.zeros((self.side_len, self.side_len), dtype=int)
        try:
            mask = Generator._draw_line(mask, genome.start_point, genome.actions)
            mask_t = Generator._draw_line(mask.T, genome.start_point, genome.actions)
            final_mask = mask | mask_t

            return GenerationResult(
                final_mask,
                TessellationType.SQUARE_TRANSLATION,
                metadata={
                    "generator_class": "GATessellationGenerator",
                    "fitness": individual.fitness,
                },
            )
        except IndexError:
            print(f"Invalid individual: {individual}, returning None")
            return None



"""Drawing module."""

from pathlib import Path
from typing import Union

import matplotlib.pyplot as plt
import numpy as np

from tessellation.procgen import GenerationResult


class Drawer:
    """Drawer class."""

    def __init__(self, cmap: str = "binary"):
        self.cmap = cmap

    def draw(self, tessellation: np.ndarray):
        """Draw the tessellation."""
        plt.imshow(tessellation, cmap=plt.get_cmap(self.cmap))

    def save_as_png(self, file_path: Union[Path, str], tessellation: np.ndarray):
        """Save the drawing as a PNG file."""
        plt.imsave(file_path, tessellation, cmap=plt.get_cmap(self.cmap))

    @staticmethod
    def tessellate(
        generation_result: GenerationResult, n_shapes: int = 5
    ) -> np.ndarray:
        """Tessellate the mask n_shapes times."""
        mask = generation_result.mask
        side_len = mask.shape[0]
        side_len_full_image = side_len * n_shapes
        tessellation = np.zeros((side_len_full_image, side_len_full_image), dtype=int)

        row_starting_color = 1
        for i in range(n_shapes):
            cell_color = row_starting_color
            for j in range(n_shapes):
                y_start = i * side_len
                y_end = y_start + side_len

                x_start = j * side_len
                x_end = x_start + side_len

                if cell_color == 0:
                    color_mask = np.logical_not(mask)
                    cell_color = 1
                else:
                    color_mask = mask
                    cell_color = 0

                tessellation[y_start:y_end, x_start:x_end] = color_mask

            row_starting_color = 1 if row_starting_color == 0 else 0

        return tessellation



"""Class that defines a LEAP problem for tessellation generation."""

from typing import Optional, Callable

import numpy as np
from leap_ec.problem import Problem

from tessellation.procgen import Action
from tessellation.procgen.ga.genome import TessellationPhenome, TessellationGenome


class TessellationProblem(Problem):
    """Class that defines a LEAP problem for tessellation generation."""

    def __init__(
        self,
        heuristic_fns: list[Callable[[TessellationPhenome], float]],
        fn_weights: Optional[list[float]] = None,
        side_len: int = 100,
    ):
        super().__init__()
        self.heuristic_fns = heuristic_fns

        if fn_weights is None:
            self.fn_weights = np.ones(len(self.heuristic_fns))
        else:
            if len(fn_weights) != len(self.heuristic_fns):
                raise ValueError(
                    "fn_weights must be exact same length as heuristic_fns if provided"
                )
            self.fn_weights = np.array(fn_weights)

        self.side_len = side_len

    def evaluate(self, phenome: TessellationPhenome, *args, **kwargs) -> np.array:
        """Evaluate the phenome using the heuristic functions."""
        heuristic_vals = [fn(phenome) for fn in self.heuristic_fns]
        return np.sum(self.fn_weights * heuristic_vals)

    def equivalent(self, first_fitness: float, second_fitness: float) -> bool:
        """Return True if the two fitness values are equivalent."""
        return first_fitness == second_fitness

    def worse_than(self, first_fitness: float, second_fitness: float) -> bool:
        """Return True if the first fitness value is worse than the second."""
        return first_fitness < second_fitness


def initialize_genome(problem: TessellationProblem):
    """Initialize a genome for the tessellation problem."""
    actions = [Action.RIGHT for _ in range(problem.side_len)]
    return TessellationGenome(actions=actions, start_point=(0, 0))
